# -*- coding: utf-8 -*-
# 
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Base classes for modeling domain objects that wrap vertices and edges.

"""

import config
from rest import Resource
from index import IndexProxy
from element import Vertex, VertexProxy, Edge, EdgeProxy
from typesystem import TypeSystem, ClassProperty

class Model(TypeSystem):
    """
    Abstract base class for Node and Relationship.

    It's a sublcass of TypeSystem, which provides a mechanism for type-checking
    database properties before they're saved in the database.

    To create a type-checked database property, use the Property class and
    datatype classes, which are located in the property module. Example::

        name = Property(String, nullable=False)
        age = Property(Integer)

    """

    # You can override this default resource.
    resource = Resource(config.DATABASE_URL)

    @property
    def eid(self):
        """Return the element ID. Override this to change it from eid."""
        return self._id
 
    #@property
    #def element_type(self):
    #    """Return the element type."""
    #    return self._data.get(config.TYPE_VAR,None)
        #return self._data[TYPE_VAR]


    @classmethod
    def get(self,_id):
        """
        Returns the element for the specified ID.

        ::param _id: The element's ID.

        """
        return self._element_proxy.get(_id)

    @classmethod
    def get_all(self):
        """Returns all the elements for the element's base class."""
        return self._element_proxy.get_all()

    @classmethod 
    def remove(self,_id,params):
        """
        Removes a property from the element for the specified ID.

        ::param _id: The element's ID.

        ::param params: The element's property to remove.

        """
        return self._element_proxy.remove(_id,params)

    @classmethod
    def create_index(self,index_keys=None,index_type="automatic"):
        """
        Creates an index for the model.

        ::param index_keys: The specific keys to index. If set to None,
                            any key can be indexed. Defaults to None.

        ::param index_type: The type of index to create. Either manual
                            or automatic. Defaults to automatic. See
                            Rexster docs for definitions.
        """
        index_name = self._get_index_name()
        index_class = self._get_index_class()
        index_attributes = (index_name,index_class,index_type,index_keys)
        return self.index_proxy.create(*index_attributes)

    @classmethod
    def delete_index(self):
        """Deletes the model's index."""
        index_name = self._get_element_key()
        return self.index_proxy.delete(index_name)

    @classmethod 
    def _get_element_proxy(self):
        """
        Returns the element's proxy class. The Node and Relationship classes
        override this.

        """
        raise NotImplementedError

    @classmethod 
    def _get_index_proxy(self):
        """Returns the index's proxy class."""
        return IndexProxy(self.resource,self)


    @classmethod
    def _get_index_name(self):
        """Returns the model's index name."""
        return self._get_element_key()

    @classmethod
    def _get_index(self):
        """Returns the model's index."""
        index_name = self._get_index_name()
        return self.index_proxy.get(index_name)

    index = ClassProperty(_get_index)
    index_proxy = ClassProperty(_get_index_proxy)


    #def element_type_sanity_check(self,results):
    #    element_type = self.get_element_type(results)
    #    if element_type is not None:
    #        if element_type != getattr(self,TYPE_VAR):
    #            raise("DB element type (%) doesn't match class element type (%s)" % \
    #                      (element_type, getattr(self,TYPE_VAR)))

    #def get_arg(self,name,default=None):
    #    return self._kwds.get(name,default)

    def _create(self,*args,**kwds):
        """
        Create a new element in the database.
        
        ::param *args: Optional dict of name/value pairs of database properties.
        ::param **kwds: name/value pairs of database properties to store.
        """
        self._set_keyword_attributes(kwds)
        self._validate_property_data()
        data = self._get_property_data()
        args = list(args)
        args.append(data)
        self.before_created()
        # Using super here b/c Vertex and Edge have different create methods
        #resp = super(self.__class__,self).create(*args,raw=True)
        resp = self._element_proxy.create(*args,raw=True)
        self._initialize_element(self.resource,resp.results)
        #self.set_element_data(resp.results)
        self.after_created()
        
    def _read(self,results):
        """
        Read an element's data that was retrieved from the DB and set its model 
        values.

        ::param results: A list containing the results returned by Rexster.

        """
        self.before_read()
        self._initialize_element(self.resource,results)
        #self.set_element_data(results)
        self._set_property_data(results)
        self.after_read()

    def _update(self,eid,kwds):
        """
        Updates the element in the database.

        ::param eid: The ID of the element to update.
        ::param **kwds: name/value pairs of database properties to store.

        """
        self.eid = eid
        self._set_keyword_attributes(kwds)
        self.save()

    def save(self):
        """
        Saves/updates the element's data in the database.
        
        """
        self._validate_property_data()
        data = self._get_property_data()        
        self.before_updated()
        #resp = super(self.__class__,self).update(self.eid,data,raw=True)
        resp = self._element_proxy.update(self.eid,data,raw=True)
        self._initialize_element(self.resource,resp.results)
        #self.set_element_data(resp.results)
        self.after_updated()

    def delete(self):
        """Deletes an element from the database."""
        # Should we make this a classmethod instead?
        # Should we provide an option to set a deleted flag or just override this?
        self.before_deleted()
        #resp = super(self.__class__,self).delete(self)
        resp = self._element_proxy.delete(self)
        self.after_deleted()
        return resp

    def initialize(self,args,kwds):
        """
        Initialize the model.

        If results is passed in, that means data was retrieved from the DB
        via a get request or gremlin query so we just set the property values
        based on what is stored in the DB.

        If eid was passed in, that means we're updating the element so we
        set the model's attributes based on the keyword variables passed in,
        and we save in the DB any attributes specified as DB Properties.

        If neither results nor eid were passed in, that means we're creating a
        new element so we set the model's attributes based on the keyword
        variables that were passed in, and we save in the DB any attributes
        specified as DB Properties.

        Also, we set self.kwds so the vars are available to the before/after wrappers.
        
        ::param *args: Optional dict of name/value pairs of database properties.

        ::param **kwds: name/value pairs of database properties to store.

        """
        # We're doing a little arg dance instead of keyword mangling.
        # This avoids problems if a user wants to use the word "results" as a
        # Property name (actually, no it doesn't, b/c you're using results
        # elsewhere. Ugh -- I hate keyword mangling.
        _id = None
        results = None
        args = list(args)

        # save kwds so it's available to before/after wrappers
        self._kwds = kwds      

        # in case someone wants to pass in a dict of data instead of by keywords
        _data = kwds.pop("_data",{})

        # in case someone passed in a dict of data plus some data by keywords
        _data.update(**kwds)
      
        if args and isinstance(args[0],Resource):
            self.resource = args.pop(0)            
        if args and isinstance(args[0],dict):
            results = args.pop(0)            
        if args and isinstance(args[0],int):
            _id = args.pop(0)            

        #print "RESULTS:", results

        self.before_initialized()
        if results is not None:
            # must have been a get or gremlin request
            Model._read(self,results)
        elif _id is not None:
            # calling Model explicitly b/c Vertex/Edge have an update method too
            Model._update(self,_id,_data)
        elif args or kwds:
            # calling Model explicitly b/c Vertex/Edge have a create method too
            Model._create(self,*args,**_data)
        else:
            # create an empty Node (can't have an empty Relationship b/c must have label)
            Model._create(self,{})
        self.after_initialized()


    def before_initialized(self):
        """Virtual method run before the model is initialized."""
        pass

    def after_initialized(self):
        """Virtual method run after the model is initialized."""
        pass

    def before_created(self):
        """Virtual method run before an element is created in the DB."""
        pass

    def after_created(self):
        """Virtual method run after an element is created in the DB."""
        pass

    def before_read(self):
        """Virtual method run before element data is read from the DB."""
        pass

    def after_read(self):
        """Virtual method run after element data is read from the DB."""
        pass

    def before_updated(self):
        """Virtual method run before an element is updated in the DB."""
        pass

    def after_updated(self):
        """Virtual method run after an element is updated in the DB."""
        pass

    def before_deleted(self):
        """Virtual method run before an element is deleted from the DB."""
        pass

    def after_deleted(self):
        """Virtual method run after an element is deleted from the DB."""
        pass
 

class Node(Vertex,Model):
    """ 
    An abstract base class used to create classes that model domain objects. It is
    not meant to be used directly

    To use this, create a subclass specific to the type of data you are 
    storing. 

    Example model declaration::

        from bulbs.model import Node
        from bulbs.property import Property, String, Integer

        class Person(Node):
            element_type = "person"

            name = Property(String, nullable=False)
            age = Property(Integer)

            def after_created():
                # include code to create relationships and to index the node
                pass

   Example usage::

        # Create a node in the DB:
        >>> james = Person(name="James Thornton")
        >>> james.eid
        3
        >>> james.name
        'James Thornton'

        # Get a node from the DB:
        >>> james = Person.get(3)

        # Update the node in the DB:
        >>> james.age = 34
        >>> james.save()

   """

    # IMPORTANT:
    # Can't do the metamagic unless object has been created so we can't use 
    # __new__ because all the initialzed vars would be on the class and not the
    # object. Don't worry about setting the object attributes for the kwds data
    # because the initialize sets them and the grandancestor Element provides 
    # DB data to you by overriding __getattr__.

    #element_type = "node"

    def __init__(self,*args,**kwds):
        """
        Initialize the node.

        ::param *args: Optional dict of name/value pairs of database properties.
        ::param **kwds: name/value pairs of database properties to store.

        """
        # pass results as the first (and only) arg to init values from gets
        # pass eid as the first arg if you are updating the element
        self.initialize(args,kwds)

    @classmethod
    def _get_element_key(self):
        """Returns the element's key that's used for stuff like the index name."""
        element_key = getattr(self,config.TYPE_VAR)
        return element_key

    @classmethod 
    def _get_index_class(self):
        """Returns the element's base class."""
        return "vertex"

    @classmethod
    def _get_element_proxy(self):
        """Returns the element's proxy."""
        return VertexProxy(self.resource,self)

    _element_proxy = ClassProperty(_get_element_proxy)

class Relationship(Edge,Model):
    """ 
    An abstract base class used to create classes that model domain objects. It is
    not meant to be used directly

    To use this, create a subclass specific to the type of data you are 
    storing. 

    Example usage for an edge between a blog entry node and its creating user::

        class CreatedBy(Relationship):
            label = "created_by"

            timestamp = Property(Float, default="current_timestamp", nullable=False)

            @property
            def entry(self):
                return Entry.get(self.outV)

            @property
            def user(self):
                return User.get(self.inV)
        
            def current_timestamp(self):
                return time.time()

          >>> entry = Entry(text="example blog entry")
          >>> james = Person(name="James")
          >>> CreatedBy(entry,james)
      
          # Or if you just want to create a basic relationship between two nodes, do::
          >>> Relationship.create(entry,"created_by",james)

    """

    def __init__(self,*args,**kwds):
        self.initialize(args,kwds)
    
    @classmethod
    def create(self,outV,label,inV,**kwds):
        """
        Create a generic relationship between two nodes.

        ::param outV: the outgoing vertex.
        ::param label: the edge's label.
        ::param inV: the incoming vertex.
        ::param **kwds: Optional keyword arguments. Name/value pairs of properties to store.
        """
        return Relationship(outV,label,inV,**kwds)
        
    @classmethod
    def _get_element_key(self):
        """Returns the element's key that's used for stuff like the index name."""
        return self.label

    @classmethod 
    def _get_index_class(self):
        """Returns the element's base class."""
        return "edge"

    @classmethod
    def _get_element_proxy(self):
        """Returns the element's proxy."""
        return EdgeProxy(self.resource,self)

    _element_proxy = ClassProperty(_get_element_proxy)




    
    
    



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#

import os
#import sys

ENVIRONMENT_VARIABLE = "BULBS_CONFIG"
DATABASE_URL = "http://localhost:8182/tinkergraph"
SAIL_URL = "http://localhost:8182/sailgraph"
TYPE_VAR = "element_type"  
MAX_RETRIES = 8
DEBUG = False

config_module = os.environ.get(ENVIRONMENT_VARIABLE,None)
#if not config_module: 
#    msg = "Settings cannot be imported because the environment variable %s is undefined." \
#        % ENVIRONMENT_VARIABLE
#    raise ImportError(msg)
try:
    mod = __import__(config_module, globals(), {}, [])
    filename = "%s/%s" % (mod.__path__[0],"bulbs_config.py")
    execfile(filename)
except Exception:
    # We catch any Exception because the user could do all sorts
    # of things wrong in the import. We can't even guarantee that
    # an ImportError comes from our __import__ call and not an import
    # attempt inside the config file without delving into the backtrace
    # using the traceback module.
    print "Unable to import your local config file. Using default settings."
    #raise




# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Interface for interacting with a graph database through Rexster.

"""
import config
from rest import Resource
from element import Vertex, VertexProxy, Edge, EdgeProxy
from index import IndexProxy
from gremlin import Gremlin

class Graph(object):
    """
    The primary interface to graph databases on the Rexster REST server.

    Instantiates the database :class:`~bulbs.rest.Resource` object using 
    the specified database URL and sets up proxy objects to the database.
        
    :keyword db_url: The URL to the specific database on Rexster. 
    :ivar vertices: :class:`~bulbs.element.VertexProxy` object for the Resource.
    :ivar edges: :class:`~bulbs.element.EdgeProxy` object for the Resource.
    :ivar indices: :class:`~bulbs.index.IndexProxy` object for the Resource.
    :ivar gremlin: :class:`~bulbs.gremlin.Gremlin` object for the Resource.

    Example::

    >>> from bulbs.graph import Graph
    >>> g = Graph()
    >>> james = g.vertices.create({'name':'James'})
    >>> julie = g.vertices.create({'name':'Julie'})
    >>> g.edges.create(james,"knows",julie)

    """

    def __init__(self,db_url=config.DATABASE_URL):
        self.resource = Resource(db_url)
        self.vertices = VertexProxy(self.resource)
        self.edges = EdgeProxy(self.resource)
        self.indices = IndexProxy(self.resource)
        self.gremlin = Gremlin(self.resource)

    #def __rshift__(self,b):
    #    return list(self)

    @property
    def V(self):
        """
        Returns all the vertices of the graph.

        Example::
        
        >>> g = Graph()
        >>> vertices = g.V

        :rtype: List of :class:`~bulbs.element.Vertex` objects. 

        """
        vertices = self.gremlin.query("g.V",Vertex,raw=True)
        return list(vertices)
    
    @property
    def E(self):
        """
        Returns all the edges of the graph. 

        Example::
        
        >>> g = Graph()
        >>> edges = g.E

        :rtype: List of :class:`~bulbs.element.Edge` objects.

        """
        edges = self.gremlin.query("g.E",Edge,raw=True)
        return list(edges)

    def idxV(self,**kwds):
        """
        Looks up a key/value pair in the vertex index and
        returns a generator containing the vertices matching the key and value.

        :keyword pair: The key/value pair to match on.
        :keyword raw: Boolean. If True, return the raw Response object. 
                      Defaults to False.

        Example::

        >>> g = Graph()
        >>> vertices = g.idxV(name="James")

        :rtype: Generator of :class:`~bulbs.element.Vertex` objects.

        You can turn the generator into a list by doing::

        >>> vertices = g.idxV(name="James")
        >>> vertices = list(vertices)

        """
        return self._idx("vertices",**kwds)
        
    #def _initialize_results(self,results,raw):
    #    if raw is True:
    #        return (result for result in results)
    #    else:
    #        return (Vertex(self.resource,result) for result in results)
        
    def idxE(self,**kwds):
        """
        Looks up a key/value pair in the edge index and 
        returns a generator containing the edges matching the key and value.

        :keyword pair: The key/value pair to match on.
        :keyword raw: Boolean. If True, return the raw Response object. 
                      Defaults to False.

        Example::

        >>> g = Graph()
        >>> edges = g.idxE(label="knows")

        :rtype: Generator of :class:`~bulbs.element.Edge` objects.

        You can turn the generator into a list by doing::

        >>> edges = g.idxE(label="knows")
        >>> edges = list(edges)

        """
        return self._idx("edges",**kwds)

    def _idx(self,index_name,**kwds):
        """
        Returns the Rexster Response object of the index look up.
        
        :param index_name: The name of the index.

        :param pair: The key/value pair to match on. 
        :keyword raw: Boolean. If True, return the raw Response object. 
                      Defaults to False.

        """
        raw = kwds.pop("raw",False)
        key, value = kwds.popitem()
        target = "%s/indices/%s" % (self.resource.db_name,index_name)
        params = dict(key=key,value=value)
        resp = self.resource.get(target,params)
        if raw:
            return resp
        if resp.results:
            class_map = dict(vertices=Vertex,edges=Edge)
            element_class = class_map[index_name]
            return (element_class(self.resource,result) for result in resp.results)

    
    def load_graphml(self,url):
        """
        Loads a GraphML file into the database, and returns the Rexster 
        response object.

        :param url: The URL of the GraphML file to load.

        """
        script = "g.loadGraphML('%s')" % url
        params = dict(script=script)
        resp = self.resource.get(self.base_target,params)
        return resp

    def save_graphml(self):
        """
        Returns a GraphML file representing the entire database.

        """

        script = """
        g.saveGraphML('data/graphml');
        new File('data/graphml').getText();
        """
        params = dict(script=script)
        resp = self.resource.get(self.base_target,params)
        return resp.results

    def clear(self):
        """
        Deletes all the elements in the graph.

        Example::

        >>> g = Graph()
        >>> g.clear()

        .. admonition:: WARNING 

           g.clear() will delete all your data!

        """
        target = self.resource.db_name
        resp = self.resource.delete(target,params=None)
        return resp


class SailGraph(Graph):
    """ An interface to for SailGraph. """

    def __init__(self,db_url=config.SAIL_URL):
        Graph.__init__(self,db_url)
        
    def add_prefix(self,prefix,namespace):
        params = dict(prefix=prefix,namespace=namespace)
        resp = self.resource.post(self._base_target(),params)
        return resp

    def get_all_prefixes(self):
        resp = self.resource.get(self._base_target(),params=None)
        return resp.results

    def get_prefix(self,prefix):
        target = "%s/%s" % (self._base_target(), prefix)
        resp = self.resource.get(target,params=None)
        return resp.results
        
    def remove_prefix(self,prefix):
        target = "%s/%s" % (self._base_target(), prefix)
        resp = self.resource.delete(target,params=None)
        return resp

    def load_rdf(self,url):
        """
        Loads an RDF file into the database, and returns the Rexster 
        response object.

        :param url: The URL of the RDF file to load.

        """
        script = "g.loadRDF('%s', 'n-triples')" % url
        params = dict(script=script)
        resp = self.resource.get(self.base_target,params)
        return resp

    def _base_target(self):
        "Returns the base target URL path for vertices on Rexster."""
        base_target = "%s/%s" % (self.resource.db_name,"prefixes")
        return base_target




import unittest

def suite():
    from rest_tests import RestTestCase
    from element_tests import VertexTestCase, VertexProxyTestCase, EdgeProxyTestCase
    from graph_tests import GraphTestCase
    from index_tests import IndexTestCase
    from model_tests import NodeTestCase, RelationshipTestCase

    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RestTestCase))
    suite.addTest(unittest.makeSuite(VertexTestCase))
    suite.addTest(unittest.makeSuite(VertexProxyTestCase))
    suite.addTest(unittest.makeSuite(EdgeProxyTestCase))
    suite.addTest(unittest.makeSuite(GraphTestCase))
    suite.addTest(unittest.makeSuite(IndexTestCase))
    suite.addTest(unittest.makeSuite(NodeTestCase))
    suite.addTest(unittest.makeSuite(RelationshipTestCase))

    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#

import config
from property import Property, convert_to_rexster, convert_to_python

TYPE_VAR = config.TYPE_VAR

# NOTE: Here the word "Property" in ClassProperty refers to a Python property (lowercase)
# We are creating a custom class b/c the standard version won't work for classmethods 
class ClassProperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

class TypeSystemMeta(type):

    def __init__(cls, name, base, namespace):
        # store the Property definitions on the class as a dictionary
        # mapping the Property name to the Property instance
        cls._properties = {}

        # loop through the class namespace looking for Property instances
        for key, value in namespace.items():
            #print key,value
            if isinstance(value, Property):
                if value.name is None:
                    # name will be none unless set via kwd param
                    value.name = key
                cls._properties[key] = value
                # now that the property reference is stored away, 
                # initialize its vars to None, the default vale (TODO), or the fget
                if value.default:
                    #TODO: Make this work for scalars too
                    fget = getattr(cls,value.default)
                    value = property(fget)
                elif value.fget:
                    # wrapped fset and fdel in str() to make the default None work with getattr
                    fget = getattr(cls,value.fget)
                    fset = getattr(cls,str(value.fset),None)
                    fdel = getattr(cls,str(value.fdel),None)
                    value = property(fget,fset,fdel)
                else:
                    value = None
                setattr(cls,key,value)

class TypeSystem(object):

    __metaclass__ = TypeSystemMeta

    def _set_property(self,key,value):
        if key in self._properties:
            datatype = type(value)
            try:
                if value:
                    python_type = self._properties[key].datatype.python_type
                    # TODO: this is a HACK until we clean up 'None' values in DB
                    if value == 'None':
                        value = None
                    value = python_type(value)

                #if not self._properties[key].is_valid(value):
                #    raise TypeError('Invalid datatype for property.')  
                    #if value:
                        # some types can't be None
                    super(TypeSystem, self).__setattr__(key, value)
            except ValueError:
                print "'%s' is not a valid value for %s, must be  %s." \
                           % (value, key, python_type)
                raise
            except AttributeError:
                print "Can't set attribute '%s' to value '%s with type'" % (key,value,datatype)
                raise
        else:
            # set value normally (it's not a database "Property")
            # But don't try to set values for get-only attributes, such as eid
            #if type(self).__dict__[key].fset is not None:
            #try:
                # TODO: make this handle properties without fset 
            super(TypeSystem, self).__setattr__(key, value)
            #except:
            #    pass

    def __setattr__(self, key, value):
        self._set_property(key,value)

    def _constructor(self, kwds):
        """A simple constructor that allows initialization from kwargs.
        
        Sets attributes on the constructed instance using the names and
        values in ``kwargs``.
        
        Only keys that are present as
        attributes of the instance's class are allowed. These could be,
        for example, any mapped columns or relationships.
        """
        cls = type(self)
        for k in kwds:
            if not hasattr(cls,k):
                raise TypeError(
                    "%r is an invalid keyword argument for %s" %
                    (k, cls.__name__))
            setattr(self, k, kwds[k])



    def _set_property_data(self,results):
        """
        Sets Property data when an element is being initialized, after it is
        retrieved from the DB.

        This will do type checking and coerce values to the right datatype.
        
        """
        
        for key, property_instance in self._properties.items():
            value = results.get(key,None)
            #if property_instance.method is None:
                # this is a normal attribute, not a derived one so set its value
            try:
                setattr(self,key,value)
            except:
                # TODO: log/warn/email regarding type mismatch
                setattr(self,key,None)

    def _get_property_data(self):
        """Returns Property data ready to be saved in the DB."""
        data = dict()
        # it should default to string in rexster so no need to 
        # convert_to_rexster()
        # should we require element_type on nodes (i.e. Model vertices)?
        if hasattr(self,TYPE_VAR):
            data[TYPE_VAR] = getattr(self,TYPE_VAR)
        for key, value in self._properties.items():
            #try:
            val = getattr(self,key)
            #if isinstance(val,Property):
            #    data[key] = None
            #else:
            data[key] = convert_to_rexster(val)
                #data[key] = val
            #except:
            #    data[key] = None
        return data

    def _validate_property_data(self):
        """
        Validates that Property data is of the right datatype before saving it
        to the DB and that the Property has a value if nullable is set to False.
        
        Call this at the top of each save() method.
        
        """
        # check that Properties with nullable is set to False 
        # actually have a value set
        for key, property_instance in self._properties.items():
            if property_instance.nullable is False:
                try: 
                    assert getattr(self,key) is not None
                except AssertionError:
                    print "Cannot set '%s' to %s: '%s' is a Property with nullable set to False" % (key, getattr(self,key), key)
                    raise

    def _set_keyword_attributes(self,kwds):
        for key, value in kwds.iteritems():
            setattr(self,key,value)




import unittest
from bulbs.model import Node, Relationship
from bulbs.property import Property, Integer, String

class Person(Node):

    element_type = "person"
    
    name = Property(String, nullable=False)
    age  = Property(Integer)

class NodeTestCase(unittest.TestCase):

    def setUp(self):
        self.james = Person(name="James", age=34)

    def test_properties(self):
        int(self.james.eid)
        assert self.james.element_type == "person"
        assert self.james.name == "James"
        assert self.james.age == 34

    def test_get(self):
        person = Person.get(self.james.eid)
        assert person == self.james
        
    def test_get_all(self):
        people = Person.get_all()
        assert len(list(people)) > 1
        
    def test_get_element_key(self):
        element_key = Person._get_element_key()
        assert element_key == "person"

    def test_index_name(self):
        index_name = Person._get_index_name()
        assert index_name == "person"

    def test_create_index(self):
        Person.delete_index()
        Person.create_index()
       
    def test_index_put_and_get(self): 
        # must test put/get together b/c self.james gets reset every time
        Person.index.put(self.james.eid,age=self.james.age)
        james = Person.index.get_unique("age",'34')
        assert self.james == james
        #Person.remove(self.james.eid,dict(age="34"))


class RelationshipTestCase(unittest.TestCase):

    def setUp(self):
        self.james = Person(name="James", age=34)
        self.julie = Person(name="Julie", age=28)
        self.relationship = Relationship.create(self.james,"knows",self.julie)
        
    def test_properties(self):
        assert self.relationship.label == "knows"
        assert self.relationship.outV._id == self.james.eid
        assert self.relationship.inV._id == self.julie.eid

def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(NodeTestCase))
    suite.addTest(unittest.makeSuite(RelationshipTestCase))

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
An interface for executing Gremlin scripts on Rexster.

"""
from config import TYPE_VAR


class Gremlin(object):
    """An interface for executing Gremlin scripts on Rexster."""

    def __init__(self,resource):
        """
        Initialize the Gremlin proxy object.
        
        :param resource: The Resource object for the database.

        """
        self.resource = resource
        self.base_target = "%s/tp/gremlin" % (self.resource.db_name)
        self.class_map = {}


    def query(self,script,*classes,**kwds):
        """
        Returns initialized results of arbitrary Gremlin scripts run through Rexster.

        :param script: Gremlin script to send to Rexster. Since this begins 
                       from the context of a graph instead of an element, the 
                       script should begin with the reference to itself 
                       (g) instead of a reference to the element (v or e). 
                       Example:

                       .. code-block:: groovy

                          // do this... 
                          g.v(1).outE('created') 
                           
                          // instead of... 
                          v.outE('created')

        :param classes: Zero or more subclasses of Element to use when 
                        initializing the the elements returned by the query. 
                        For example, if Person is a subclass of Node (which 
                        is defined in model.py and is a subclass of Vertex), 
                        and the query returns person elements, pass in the 
                        Person class and the method will use the element_type
                        defined in the class to initialize the returned items
                        to a Person object.

        :param return_keys: Optional keyword param. A comman-separated list of
                            keys (DB properties) to return. If set to None, it
                            returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 
        
        """
        raw = kwds.pop('raw',False)
        resp = self._query(script,*classes,**kwds)
        for result in resp.results:
            yield self._initialize_result(result,raw)
 

    def execute(self,script,**kwds):
        """
        Returns raw results of arbitrary Gremlin scripts run through Rexster.

        :param script: Gremlin script to send to Rexster. Since this begins 
                       from the context of a graph instead of an element, the 
                       script should begin with the reference to itself 
                       (g) instead of a reference to the element (v or e). 
                       Example:

                       .. code-block:: groovy

                          // do this... 
                          g.v(1).outE('created') 
                           
                          // instead of... 
                          v.outE('created')

        :param return_keys: Optional keyword param. A comman-separated list of
                            keys (DB properties) to return. If set to None, it
                            returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 
        
        """
        raw = kwds.pop('raw',False)
        classes = ()
        resp = self._query(script,*classes,**kwds)
        if raw:
            return resp
        return list(resp.results)


    def lds_query(self,script,raw=False):
        """Return queries on remote LinkedData stores."""
        # this hack will go away when the LinkedDataSail mods to Rexster are finished. 
        lds_hack = "g = new LinkedDataSailGraph(new MemoryStoreSailGraph());"
        script = lds_hack + self._trim_lines(script)
        params = dict(script=script)
        resp = self.resource.get(self.base_target,params)
        for result in resp.results:
            yield self._initialize_result(result,raw)

    def _query(self,script,*classes,**kwds):
        """Returns raw results of arbitrary Gremlin scripts run through Rexster."""
        script = self._trim_lines(script)
        # do these imports here to avoid circular dependency issue with element.py
        from element import Vertex, Edge
        default_class_map = dict(vertex=Vertex,edge=Edge)
        self.class_map.update(default_class_map)
        self._add_classes(classes)
        return_keys = kwds.pop('return_keys',None)
        params = dict(script=script)
        if return_keys:
            params.update(return_keys=return_keys)
        resp = self.resource.get(self.base_target,params)
        return resp

    def _element_query(self,element,script,*classes,**kwds):
        """
        This is called by Element and probably won't be useful anywhere else.

        Returns a generator containing the results of the Gremlin script. 
        Remember you can always use the list() function to turn an iterator or 
        a generator into a list. Sometimes it's useful to turn a generator into
        a list when doing unittests or when you want to check how many items 
        are in the results. For example::
            
            results = person.gremlin('v.outE.inV')
            results = list(results)
            
        :param script: Gremlin script to send to Rexster. Since this begins 
                       from the context of an element instead of a graph, the 
                       script should begin with the reference to itself 
                       (v or e) instead of a reference to the graph (g). 
                       Example:

                       .. code-block:: groovy

                       // do this... 
                       v.outE('created') 
                        
                       // instead of... 
                       g.v(1).outE('created')

        :param classes: Zero or more subclasses of Element to use when 
                        initializing the the elements returned by the query. 
                        For example, if Person is a subclass of Node (which 
                        is defined in model.py and is a subclass of Vertex), 
                        and the query returns person elements, pass in the 
                        Person class and the method will use the element_type
                        defined in the class to initialize the returned items
                        to a Person object.

        :param kwds: name/value pairs of optional arguments:
                        
        :param return_keys: Optional keyword param. A comman-separated list of
                            keys (DB properties) to return. If set to None, it
                            returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 

        """
        script = self._trim_lines(script)
        default_class_map = kwds.pop('default_class_map')
        self.class_map.update(default_class_map)
        return_keys = kwds.pop('return_keys',None)
        raw = kwds.pop('raw',False)
        self._add_classes(classes)
        #path = element.path()
        target = "%s/%s/tp/gremlin" % (element._base_target(),element._id)
        params = dict(script=script)
        if return_keys:
            params.update(return_keys=return_keys)
        print "TARGET", target
        resp = self.resource.get(target,params)
        for result in resp.results:
            yield self._initialize_result(result,raw)

    def _trim_lines(self,script):
        """
        Returns a clean Gremlin script, trimmed of its whitespace and 
        ready to be sent Rexster.

        :param script: The Gremlin script you want to clean up.

        """ 
        lines = []
        for line in script.split('\n'):
            line = line.strip()
            if line:
                lines.append(line)
        script = ";".join(lines)
        return script
        
    def _initialize_result(self,result,raw):
        """
        Maybe initialize the object returned in the query result.

        If the element's class was passed in to the query method and raw is False, 
        it will initialize the element. If raw is True, it won't try to initialize
        the element and will just return the raw result.

        :param result: An individual result item returned by Rexster (not a result list).

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 

        """
        if raw is False:
            element_class = self._get_element_class(result)
            if element_class:
                result = element_class(self.resource,result)
        return result
       
    def _add_classes(self,classes):
        """Add a list of element classes to the class_map."""
        for element_class in classes:
            self._add_element_class(element_class)   
 
    def _add_element_class(self,element_class):
        """Ad an individual element class to the class map."""
        element_type = getattr(element_class,TYPE_VAR,None)
        if element_type:
            self.class_map[element_type] = element_class

    def _get_element_class(self,result):
        """
        Return the element class for a query result item.

        :param result: An individual result item returned by Rexster (not a result list).

        """
        element_type = result.get(TYPE_VAR,None)
        if element_type not in self.class_map:
            element_type = result['_type']
        return self.class_map.get(element_type,None)
 



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
An interface for interacting with indices on Rexster.

"""

from element import Vertex, Edge
from gremlin import Gremlin
from config import TYPE_VAR

class Index(object):
    """
    Creates, retrieves, and deletes indices provided by the graph database.

    Use this class to get, put, and update items in an index.
    

    :param resource: The Resource object for the database.

    :param results: The results list returned by Rexster.

    :param classes: Zero or more subclasses of Element to use when 
                    initializing the the elements returned by the query. 
                    For example, if Person is a subclass of Node (which 
                    is defined in model.py and is a subclass of Vertex), 
                    and the query returns person elements, pass in the 
                    Person class and the method will use the element_type
                    defined in the class to initialize the returned items
                    to a Person object.

    Example that creates an index for Web page URL stubs, 
    adds an page element to it, and then retrieves it from the index::

        >>> graph = Graph()
        >>> graph.indices.create("page","vertex","automatic","[stub]")
        >>> index = graph.indices.get("page")
        >>> index.put("stub",stub,page._id)
        >>> page = index.get("stub",stub)

    """

    def __init__(self,resource,results,*classes):
        """
        Initializes the Index object.
        """
        self.resource = resource
        self._results = results
        self._classes = classes
        self._class_map = dict(vertex=Vertex,edge=Edge)
        self._update_class_map(classes)

    @property
    def index_name(self):
        """Returns the index name."""
        return self._results['name']

    @property
    def index_class(self):
        """Returns the index class, which will either be vertex or edge."""
        return self._results['class']
        # Temp hack until we fix Rexster so it doesn't return the full Java 
        # WE don't need this anymore. Stephen fixed the Rexster bug! :)
        # class
        #return self._get_index_class(index_class)

    @property
    def index_type(self):
        """Returns the index type, which will either be automatic or manual."""
        return self._results['type']
    
    def put(self,_id,key=None,value=None,**pair):
        """
        Put an element into the index at key/value and return Rexster's 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.

        """
        # NOTE: if you ever change the _id arg to element, change remove() too
        if pair:
            key, value = pair.popitem()
        target = "%s/%s" % (self._base_target(), self.index_name)
        params = {'key':key,'value':value,'class':self.index_class,'id':_id}
        resp = self.resource.post(target,params)
        return resp

    def put_unique(self,_id,key=None,value=None,**pair):
        """
        Put an element into the index at key/value and overwrite it if an 
        element already exists at that key and value; thus, there will be a max
        of 1 element returned for that key/value pair. Return Rexster's 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in 
                       the form of name='James'.

        """
        for result in self.get(key,value,**pair):
            self.remove(result._id,key,value,**pair)
        resp = self.put(_id,key,value,**pair)
        return resp

    def update(self,_id,key=None,value=None,**pair):
        """
        Update the element ID for the key and value and return Rexsters' 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.
        """
        return self.put_unique(_id,key,value,**pair)

    def get(self,key=None,value=None,raw=False,**pair):
        """
        Return a generator containing all the elements with key property equal 
        to value in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize the results. Defaults to False. 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.
        """
        if pair:
            key, value = pair.popitem()
        target = "%s/%s" % (self._base_target(),self.index_name)
        params = dict(key=key,value=value)
        resp = self.resource.get(target,params)
        for result in resp.results:
            yield self._initialize_result(result,raw)
         
    def get_unique(self,key=None,value=None,raw=False,**pair):
        """
        Returns a max of 1 elements matching the key/value pair in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize the results. Defaults to False. 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.
        """

        generator = self.get(key,value,raw,**pair)
        results = list(generator)
        if len(results) > 1:
            raise("Unique index result contains more than one item.")
        if results:
            return results[0]

    def count(self,key=None,value=None,**pair):
        """
        Return a count of all elements with 'key' equal to 'value' in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.
        """
        if pair:
            key, value = pair.popitem()
        target = "%s/%s/count" % (self._base_target(),self.index_name)
        params = dict(key=key,value=value)
        resp = self.resource.get(target,params)
        return resp.total_size
     
    def keys(self):
        """Return the index's keys."""
        target = "%s/%s/keys" % (self._base_target(),self.index_name)
        resp = self.resource.get(target,params=None)
        return list(resp.results)

    def remove(self,_id,key=None,value=None,**pair):
        """
        Remove the element from the index located by key/value.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param **pair: Optional keyword param. Instead of supplying key=name 
                       and value = 'James', you can supply a key/value pair in
                       the form of name='James'.
        """

        if pair:
            key, value = pair.popitem()
        target = "%s/%s" % (self._base_target(),self.index_name)
        params = {'key':key,'value':value,'class':self.index_class,'id':_id}
        resp = self.resource.delete(target,params)
        return resp

    def rebuild(self,raw=False):
        # need class_map b/c the Blueprints need capitalized class names, 
        # but Rexster returns lower-case class names for index_class
        class_map = dict(vertex='Vertex',edge='Edge')
        klass = class_map[self.index_class]
        script = "index = g.getIndex('%s',%s);" % (self.index_name,klass)
        script = script + "AutomaticIndexHelper.reIndexElements(g, index, g.getVertices())"
        resp = Gremlin(self.resource).execute(script,raw=True)
        if raw or not resp.results:
            return resp
        return list(resp.results)

    def _get_index_class(self,index_class):
        """
        Returns the index class. This is a hack that can go away when Rexster
        bug is fixed.

        :param index_class: The index class returned by Rexster. Example: 
                            "class":"com.tinkerpop.blueprints.pgm.Vertex"
        """

        if "Vertex" in index_class:
            index_class = "vertex"
        elif "Edge" in index_class:
            index_class = "edge"
        else:
            raise Exception("Invalid index class")
        return index_class

    def _update_class_map(self,classes):
        """
        Adds a list of element classes to the class_map.

        :param classes: A list of element classes to add to the class_map.
        """
        for element_class in classes:
            self._add_element_class(element_class)
            
    def _base_target(self):
        """Returns the base URL for the indices resource on Rexster.""" 
        base_target = "%s/indices" % (self.resource.db_name)
        return base_target

    ################ 3 similar methods are in gremlin.Gremlin() #########
    def _initialize_result(self,result,raw):
        """
        Maybe initialize the object returned in the query result.

        If the element's class was passed in to the query method and raw is 
        False, it will initialize the element. If raw is True, it won't try to 
        initialize the element and will just return the raw result.

        :param result: An individual result item returned by Rexster (not a 
                       result list).

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 

        """
        if raw is False:
            element_class = self._get_element_class(result)
            if element_class:
                result = element_class(self.resource,result)
        return result

    def _add_element_class(self,element_class):
        """Ad an individual element class to the class map."""
        element_type = getattr(element_class,TYPE_VAR,None)
        if element_type:
            self._class_map[element_type] = element_class

    def _get_element_class(self,result):
        """
        Return the element class for a query result item.

        :param result: An individual result item returned by Rexster (not a  
                       result list).

        """
        if len(self._classes) == 1:
            # if you pass in just one class, we assume you know what you're 
            # doing and you are saying that all of the results will be of the 
            # class you specified, even though the results don't have an 
            # element_type property saved in the DB
            # TODO: Maybe upgrade the Gremlin method to behave the same way.
            element_class = self._classes[0]
        else:
            element_type = result.get(TYPE_VAR,None)
            if element_type not in self._class_map:  
                # just return a generic Vertex/Edge class
                element_type = result['_type']
            element_class = self._class_map.get(element_type,None) 
        return element_class
    ##########################################################################


class IndexProxy(object):
    """
    An interface for interacting with indices on Rexster.

    :param resource: The Resource object for the database.

    :param classes: Zero or more subclasses of Element to use when 
                    initializing the the elements returned by the query. 
                    For example, if Person is a subclass of Node (which 
                    is defined in model.py and is a subclass of Vertex), 
                    and the query returns person elements, pass in the 
                    Person class and the method will use the element_type
                    defined in the class to initialize the returned items
                    to a Person object.
    """
    
    def __init__(self,resource,*classes):
        """
        Initialize the Gremlin proxy object.

        """

        self.resource = resource
        self._classes = classes
                        
    def create(self,index_name,index_class,index_type,index_keys=None):
        """ 
        Adds an index to the database and returns it. 

        index_keys must be a string in this format: '[k1,k2]'
        Don't pass actual list b/c keys get double quoted.

        :param index_name: The name of the index to create.

        :param index_class: The class of the elements stored in the index. 
                            Either vertex or edge.
        
        """
        target = "%s/%s" % (self._base_target(), index_name)
        params = {'class':index_class,'type':index_type}
        if index_keys: 
            params.update({'keys':index_keys})
        resp = self.resource.post(target,params)
        return Index(self.resource,resp.results,*self._classes)
        
    def get(self,index_name):
        """
        Returns the Index object with the specified name.

        :param index_name: The name of the index.

        """
        target = "%s/%s" % (self._base_target(), index_name)
        resp = self.resource.get(target,params=None)
        # NOTE: Get returns the index properties as part of content,
        # not results. I've filed a bug report. Stephen fixed it :)
        #results = {'name':resp.index_name,
        #           'type':resp.index_type,
        #           'class':resp.index_class}
        return Index(self.resource,resp.results,*self._classes)

    def delete(self,name):
        """
        Deletes/drops an index and returns the Rexster Response object.

        :param name: The name of the index.

        """
        target = "%s/%s" % (self._base_target(),name)
        resp = self.resource.delete(target,params=None)
        return resp

    def _base_target(self):
        """Returns the base URL target for the indices resource on Rexster."""
        base_target = "%s/indices" % (self.resource.db_name)
        return base_target



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Low-level module for connecting to the Rexster REST server and
returning a Response object.

"""

import urllib
import urllib2
import httplib2
import simplejson as json
from urlparse import urlsplit
from posixpath import basename
import config

DEBUG = config.DEBUG
MAX_RETRIES = config.MAX_RETRIES

class Resource(object):
    """Used for connecting to the Rexster REST server."""

    def __init__(self,db_url):
        """
        Initializes a resource object.

        :param db_url: the base URL of Rexster.

        """
        self.db_url = db_url
        if db_url.endswith("/"):
            # strip off trailing slash
            db_url = db_url[:-1]
        url_object = urlsplit(db_url)
        self.base_url = "%s://%s" % (url_object.scheme,url_object.netloc)
        self.db_name = basename(url_object.path)        

    def get(self,target,params):
        """Convenience method that sends GET requests to the resource.""" 
        return self.request("GET",target,params)
       
    def post(self,target,params):
        """Convenience method that sends POST requests to the resource."""
        return self.request("POST",target,params)
    
    def delete(self,target,params):
        """Convenience method that sends DELETE requests to the resource."""
        return self.request("DELETE",target,params)
    
    def request(self, method, target, params):
        """
        Sends requests to the resource.

        :param method: either GET, POST, or DELETE.
        :param target: the URL path relative to the database URL you specified 
                       in either config.py or that you passed in as an argument
                       when you instantiated the resource.
        :param params: a dict of query-string parameters to include in the URL 
        """

        assert method in ("GET","POST","DELETE"), \
            "Only GET, POST, DELETE are allowed."

        headers = {'Accept': 'application/json', 'connection': 'close'}

        # httplib2 let's you cache http responses with memcache 
        # really cool, look into that later
        # http = httplib2.Http(memcache)

        # the indicies condition is a hack until you implement 
        # list type for indices keys
        if params and method is "POST" and "indices" not in target:
            url = self._build_url(method, target, params=None)
            body = urllib.urlencode(params)
            post_headers = {'Content-type': 'application/x-www-form-urlencoded'}
            headers.update(post_headers)
        else:
            url = self._build_url(method, target, params)
            body = None
                
        if DEBUG:
            print "REST url:  ", url
            print "REST body: ", body

        attempt = 0
        ok = False
        http = httplib2.Http()       
        while not ok and (attempt < MAX_RETRIES):
            resp = Response(http.request(url, method, body, headers))
            attempt += 1
            ok = resp.ok            

        # TODO: check http_status for 401s
        #assert resp.ok == True

        # TODO: if resp.http_status == 200, 
        # else (where to put this?...maybe better at a higher level)

        return resp
    
    def _build_url(self, method, target, params):
        """Used internally to construct the complete URL to the service."""
        if type(target) == unicode:
            target = target.encode("utf8")
        url = "%s/%s" % (self.base_url, urllib2.quote(target)) 
        if params:            
            url = "%s?%s" % (url, urllib.urlencode(params))
        return url
            
        
class Response(object):
    """A container for the Rexster HTTP response."""

    def __init__(self, http_resp):
        """
        NOTE: headers is a httplib2 Response object, content is a string
        see http://httplib2.googlecode.com/hg/doc/html/libhttplib2.html
        """
        headers, content = http_resp
        self._response_handler(headers)
        if DEBUG:
            # only set raw if DEBUG=True to save on memory
            self.raw = http_resp
        try:
            self.headers = headers
            # don't save content on the resp object else you'll bloat it
            content = json.loads(content)
            self.ok = True

            # true if he response was returned from the cache
            self.fromcache = headers.fromcache
            
            # version of HTTP that the server supports. A value of 11 means 1.1
            self.http_version = headers.version
            
            # numerical HTTP status code returned in the response
            self.http_status = headers.status
            
            # human readable component of the HTTP response status code
            self.http_reason = headers.reason

            # previous response object if redirects are followed
            self.http_previous = headers.previous
        
            #
            # header values not set in resp object but stored in the dict
            #
            # URI that was ultimately requested
            self.content_location = self.headers.get('content-location')
            self.tansfer_encoding = self.headers.get('transfer-encoding')
            self.server = self.headers.get('server')
            self.date = self.headers.get('date')
            self.allow_origin = self.headers.get('access-control-allow-origin')
            self.content_type = self.headers.get('content-type')
            
            #
            # body values always returned by reXster
            #            
            self.rexster_version = content.get('version')
            self.query_time = content.get('queryTime')
            if self.query_time:
                self.query_time = float(self.query_time)
 
            #
            # body values sometimes returned, depending on resource requested
            #
            # if results is returned, it's a list of dicts, and
            # the dicts can contain any number of key/value pairs depending on 
            # what is stored in each item -- items can be a node,edge...
            self.success = bool(content.get('success'))
            if "results" in content and type(content['results']) == list:
                self.results = iter(content['results'])
            else:
                self.results = content.get('results')
            self.index_name = content.get('name')
            self.index_class = content.get('class')
            self.index_type = content.get('type')            
            self.total_size = content.get('totalSize')

        except ValueError:
            # looks like this isn't json, data is None
            self.ok = False

    def _response_handler(self,headers):
        """TODO: handle response errors.""" 
        # Is there a list of error codes Rexster sends?
        # Putting this in element won't work b/c element doesn't get the full 
        # resp anymore, just results. Prob relegate to an resp/error helper
        pass




import unittest
from bulbs import config
from bulbs.rest import Resource


class RestTestCase(unittest.TestCase):
    
    def setUp(self):
        self.resource = Resource(config.DATABASE_URL)

    def test_post(self):
        name_in = "james"
        email_in = "james@jamesthornton.com"

        target = "%s/%s" % (self.resource.db_name, "vertices")
        params = dict(name=name_in, email=email_in)
        resp = self.resource.post(target,params)
                
        oid = int(resp.results['_id'])
        element_type = resp.results['_type']
        name_out = resp.results['name']
        email_out = resp.results['email']
        
        assert type(resp.results) == dict
        assert type(oid) == int
        assert element_type == 'vertex'
        assert name_in == name_out 
        assert email_in == email_out

        # use the results of this function for get and delete tests 
        return resp
        
    def test_get(self):
        resp1 = self.test_post()
        oid1 = int(resp1.results['_id'])
        element_type1 = resp1.results['_type']
        name1 = resp1.results['name']
        email1 = resp1.results['email']

        target = "%s/%s/%d" % (self.resource.db_name, "vertices", oid1)
        params = dict(name=name1, email=email1)
        
        resp2 = self.resource.get(target,params)
        oid2 = int(resp2.results['_id'])
        element_type2 = resp2.results['_type']
        name2 = resp2.results['name']
        email2 = resp2.results['email']

        assert oid1 == oid2
        assert element_type1 == element_type2
        assert name1 == name2
        assert email1 == email2

    def test_delete(self):

        resp1 = self.test_post()
        oid1 = int(resp1.results['_id'])
        name1 = resp1.results['name']
        email1 = resp1.results['email']

        # make sure it's there
        assert type(oid1) == int

        # delete it
        target = "%s/%s/%d" % (self.resource.db_name, "vertices", oid1)
        params = dict(name=name1, email=email1)

        resp2 = self.resource.delete(target,params)

        # verify it's gone
        assert "SNAPSHOT" in resp2.rexster_version
        assert type(resp2.query_time) == float
        assert resp2.results == None


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RestTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#

def xstr(s):
    return '' if s is None else str(s)

def quote(s):
    s = xstr(s)
    if s:
        s = "'%s'" % (s)
    return s

def initialize_element(class_map,key,result):
    element_class = class_map[key]
    return element_class(result)
 
def element_repr(element):
    if element._type == "vertex":
        return "==>v[%d]" % (element.eid)
    elif element._type == "edge":
        return "==>e[%d][%d-%s->%d]" % \
            (element.eid,element.outV.eid,element.label,element.inV.eid)

def coerce_id(_id):
    # try to coerce the element ID string to an int.
    # ORIENTDB USES STRINGS SO THIS WON'T WORK FOR IT
    try:
        return int(_id)
    except:
        return _id


#def pop_arg(self,args,position,arg_type):
#    value = None
#    args = list(args)
#    if args and isinstance(args[position], arg_type):
#        value = args.pop(position)            
#    return args, value






# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Vertex and Edge container classes and associated proxy classes to interact with
Rexster.

"""
import utils
from gremlin import Gremlin

class Element(object):
    """This is an abstract base class for Vertex and Edge"""
    
    def __init__(self,resource,results):
        """
        Initializes an element after it is added, updated, or retrieved from 
        the database.

        :param resource: The Resource object for the database.

        :param results: The results list returned by Rexster.
        
        """
        # NOTE: Put all the init stuff in initialize_element() because 
        # Model() calls it explicitly instead of super().__init__()
        self._initialize_element(resource,results)

    def _initialize_element(self,resource,results):
        """
        Initialize the element's resource, _data, and Gremlin proxy. This is
        called explicitly by Element.__init__ and Model.__init__

        :param resource: The Resource object for the database.

        :param results: The results list returned by Rexster.

        """
        self.resource = resource
        self._set_element_data(results)
        self._gremlin = Gremlin(self.resource)

    def _set_element_data(self,results):
        """ 
        Set the elements data returned by the DB.

        :param results: The results list returned by Rexster.
     
        """
        self._data = results
 
    def __getattr__(self,attribute):
        """
        Returns the value stored in the DB if the property hasn't been 
        set explicitly. 

        If you explicitly set/change the values of an element's properties,
        make sure you call save() to updated the values in the DB.
        
        """
        try:
            return self._data[attribute]
        except:
            raise AttributeError("%s not found and not in self._data" \
                                     % (attribute))

    def __len__(self):
        """Returns the number of items stored in the DB results"""
        return len(self._data)

    def __contains__(self, item):
        """Returns True if attribute is a key that has been stored in the DB"""
        return item in self._data

    def __eq__(self, obj):
        """Returns True if the elements are equal"""
        return (hasattr(obj, "_id")
                and self._id == obj._id
                and hasattr(obj, "_data")
                and self._data == obj._data
                and hasattr(obj, "__class__")
                and self.__class__ == obj.__class__
                )

    def __ne__(self, obj):
        """Returns True if the elements are not equal."""
        return not self.__eq__(obj)

    def __repr__(self):
        """Returns the string representation of the attribute."""
        return self.__unicode__()
    
    def __str__(self):
        """Returns the string representation of the attribute."""
        return self.__unicode__()
    
    def __unicode__(self):
        """Returns the unicode representation of the attribute."""
        #return u"<Rexster %s %s: %s>" % \
        #    (self.resource.db_name, self.__class__.__name__, self._id)
        return u"<%s: %s>" % (self.__class__.__name__,
                              self._proxy(self.resource)._uri(self._id))

    @property
    def _id(self):
        """
        Returns the element ID. This is the element's "primary key"; however,
        some DBs (such as neo4j) reuse IDs if they are deleted so 
        be careful with how you use them. If you want to guarantee
        they are unique across the DB's lifetime either don't 
        physically delete elements and just set a deleted flag, or
        use some other mechanism, such as an external sequence or 
        a hash.
        
        """
        #int(self._data['_id'])
        return utils.coerce_id(self._data['_id'])

    @property
    def _type(self):
        """Returns the _type set by Rexster: either vertex, edge, or index."""
        return self._data['_type']

    @property
    def _proxy(self):
        """Returns the element's proxy to Rexster."""
        proxy_map = dict(vertex=VertexProxy,edge=EdgeProxy)
        return proxy_map[self._type]

    @property
    def map(self):
        """Returns a dict of the element's data that's stored in the DB."""
        private_keys = ['_id','_type','_outV','_inV','_label']
        map_ = dict()
        for key, value in self._data.items():
            if key not in private_keys:
                map_.update({key:value})
        return map_

    def gremlin(self,script,*classes,**kwds):
        """
        Returns a generator containing the results of the Gremlin script. 
        Remember you can always use the list() function to turn an iterator or 
        a generator into a list. Sometimes it's useful to turn a generator into
        a list when doing unittests or when you want to check how many items 
        are in the results.
       
        :param script: Gremlin script to send to Rexster. Since this begins 
                       from the context of an element instead of a graph, the 
                       script should begin with the reference to itself 
                       (v or e) instead of a reference to the graph (g). 
                       Example:

                       .. code-block:: groovy

                       // do this... 
                       v.outE('created') 
                        
                       // instead of... 
                       g.v(1).outE('created')
        :param classes: Zero or more subclasses of Element to use when 
                          initializing the the elements returned by the query. 
                          For example, if Person is a subclass of Node (which 
                          is defined in model.py and is a subclass of Vertex), 
                          and the query returns person elements, pass in the 
                          Person class and the method will use the element_type
                          defined in the class to initialize the returned items
                          to a Person object.
        :keyword return_keys: Optional keyword param. A comma-separated list of
                              keys (DB properties) to return. If set to None, it
                              returns all properties. Defaults to None.
        :keyword raw: Optional keyword param. If set to True, it won't try to 
                      initialize data. Defaults to False. 

        :rtype: Generator of items. The data types of the items returned vary 
                depending on the query.

        Example::
       
        >>> from bulbs.graph import Graph()
        >>> g = Graph()
        >>> james = g.vertices.get(3) 
        >>> script = "v.outE('knows').inV"
        >>> results = james.gremlin(script)


        """
        # TODO: now that we're using a proxy, should we always include the 
        # calling class in the class_map?
        return_keys = kwds.pop('return_keys',None)
        raw = kwds.pop('raw',False)
        class_map = dict(vertex=Vertex,edge=Edge)
        kwds = dict(default_class_map=class_map,return_keys=return_keys,raw=raw)
        return self._gremlin._element_query(self,script,*classes,**kwds)


class Vertex(Element):
    """
    A container for Vertex elements returned by Rexster.

    :param resource: The Resource object for the database.
    :param results: The element data returned by Rexster.

    Example::
        
    >>> from bulbs.graph import Graph
    >>> from whybase.person import Person
    >>> g = Graph()
    >>> g.inV(Person,label="knows")

    """     
  
            
    def outE(self,*args,**kwds):
        """
        Return the outgoing edges of the vertex.

        :arg classes: Zero or more classes to used for initializing the 
                        objects returned in the query results.

        :keyword label: Optional. The edge label. Defaults to None.
                        
        :keyword return_keys: Optional. A comman-separated list of
                              keys (DB properties) to return. If set to None, it
                              returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize data. Defaults to False. 

        """
        label, classes = self._parse_args(*args)
        script = "v.outE(%s)" % utils.quote(label)
        return self.gremlin(script,*classes,**kwds)

    def inE(self,*args,**kwds):
        """
        Return the incoming edges of the vertex.


        :param classes: Zero or more classes to used for initializing the 
                          objects returned in the query results.

        :param kwds: name/value pairs of optional arguments:

        :param label: Optional keyword param. The edge label.
                        
        :param return_keys: Optional keyword param. A comman-separated list of
                             keys (DB properties) to return. If set to None, it
                             returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                     initialize data. Defaults to False. 
        """
        label, classes = self._parse_args(*args)
        script = "v.inE(%s)" % utils.quote(label)
        return self.gremlin(script,*classes,**kwds)

    def bothE(self,*args,**kwds):
        """
        Return all incoming and outgoing edges of the vertex.

        :param classes: Zero or more classes to used for initializing the 
                         objects returned in the query results.

        :param kwds: name/value pairs of optional arguments:

        :param label: Optional keyword param. The edge label.
                        
        :param return_keys: Optional keyword param. A comman-separated list of
                             keys (DB properties) to return. If set to None, it
                             returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                     initialize data. Defaults to False. 
        """
        label, classes = self._parse_args(*args)
        script = "v.bothE(%s)" % utils.quote(label)
        return self.gremlin(script,*classes,**kwds)

    def outV(self,*args,**kwds):
        """ 
        Return the out-adjacent vertices to the vertex.

        :param classes: Zero or more classes to used for initializing the 
                        objects returned in the query results.

        :param kwds: name/value pairs of optional arguments:

        :param label: Optional keyword param. The edge label.
                        
        :param return_keys: Optional keyword param. A comman-separated list of
                             keys (DB properties) to return. If set to None, it
                             returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                     initialize data. Defaults to False. 
        """
        label, classes = self._parse_args(*args)
        script = "v.out(%s)" % utils.quote(label)
        return self.gremlin(script,*args,**kwds)
        
    def inV(self,*args,**kwds):
        """
        Return the in-adjacent vertices of the vertex.

        :param classes: Zero or more classes to used for initializing the 
                        objects returned in the query results.

        :param kwds: name/value pairs of optional arguments:

        :param label: Optional keyword param. The edge label.
                        
        :param return_keys: Optional keyword param. A comman-separated list of
                            keys (DB properties) to return. If set to None, it
                            returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                     initialize data. Defaults to False. 
        """
        label, classes = self._parse_args(*args)
        script = "v.in(%s)" % utils.quote(label)
        return self.gremlin(script,*classes,**kwds)
        
    def bothV(self,*args,**kwds):
        """
        Return all incoming- and outgoing-adjacent vertices of vertex.

        :param label: Optional, first positional arg. The edge label.

        :param classes: Optional, first or second positional arg.
                        Zero or more classes to used for initializing the 
                        objects returned in the query results.

        :param kwds: name/value pairs of optional arguments:


                        
        :param return_keys: Optional keyword param. A comman-separated list of
                             keys (DB properties) to return. If set to None, it
                             returns all properties. Defaults to None.

        :param raw: Optional keyword param. If set to True, it won't try to 
                     initialize data. Defaults to False. 
        """ 
        label, classes = self._parse_args(*args)
        script = "v.both(%s)" % utils.quote(label)
        return self.gremlin(script,*classes,**kwds)
            
    def get_edges(self,direction,label):
        """
        Returns a generator containing the the edges attached to the vertex 
        that have the specified label and direction. This is a native Rexster 
        query and does not go through Gremlin, and because it doesn't have the 
        compile-time overhead that Gremlin does, it's somewhat faster.

        However, it does not a mechanism for initializing edges to Edge
        subclasses -- everything is returned as a generic Edge object.
        
        :param direction: The direction of edges: Either inE, outE, or bothE.

        :param label: The edges' label. 

        """
        assert direction in ('inE','outE','bothE')
        target = "%s/%s/%s" % (self._base_target(),self._id, direction)
        params = None
        if label: 
            params = dict(_label=label)
        resp = self.resource.get(target,params)           
        for result in resp.results:
            yield Edge(result)

    def _parse_args(self,*args):
        args = list(args)
        label = None
        classes = []
        if args:
            if isinstance(args[0], basestring):
                label = args.pop(0)
        if args:
            classes = args
        return label, classes
    
    def _base_target(self):
        "Returns the base target URL path for vertices on Rexster."""
        base_target = "%s/%s" % (self.resource.db_name,"vertices")
        return base_target

        
class Edge(Element):
    """A container for Edge elements returned by Rexster."""

    @property
    def _outV(self):
        """Returns the outgoing vertex ID of the edge."""
        #return int(self._data['_outV'])
        return utils.coerce_id(self._data['_outV'])
        
    @property
    def _inV(self):
        """Returns the incoming vertex ID of the edge."""
        #return int(self._data['_inV'])
        return utils.coerce_id(self._data['_inV'])

    # TODO: Make outV and inV return a specific Model class and not a generic 
    # Vertex
    @property
    def outV(self):
        """Returns the outgoing Vertex of the edge."""
        return VertexProxy(self.resource).get(self._outV)
    
    @property
    def inV(self):
        """Returns the incoming Vertex of the edge."""
        return VertexProxy(self.resource).get(self._inV)
    
    @property
    def label(self):
        """Returns the edge's label."""
        return self._label


class ElementProxy(object):
    """An abstract base class for VertexProxy and EdgeProxy."""
    
    def __init__(self,resource,element_class):
        """
        Initializes the proxy object.

        :param resource: The Resource object for the database.

        :param element_class: The container class for the element type
                               used to instantiate the element object. 

        """
        self.resource = resource
        self.element_class = element_class

    def _base_target(self):
        base_target = "%s/%s" % (self.resource.db_name,self._path())
        return base_target

    def _uri(self,_id):
        """
        Returns the URI of the element.

        :param _id: The element ID.

        """
        uri = "%s/%s/%s" % (self.resource.db_url,self._path(),_id)
        return uri

    def create(self,data,raw=False):
        """ 
        Adds an element to the database and returns it. 

        :param data: The element's property data to store.

        :param raw: Optional keyword param. If raw is False (which is the 
                     default) it will try to instantiated the element to the 
                     element_class specified when the proxy object was created.
                     However, if the raw param is set to True, it will won't 
                     instantiate the object and will return the raw Response 
                     object.
        """
        target = self._base_target()
        resp = self.resource.post(target,data)
        if raw is True:
            return resp
        # TODO: implement a version for model that returns self()
        return self.element_class(self.resource,resp.results)

    def update(self,_id,data,raw=False):
        """ 
        Updates a vertex in the graph DB and returns it. 
        
        :param _id: The element ID.

        :param data: The element's property data to store.

        :param raw: Optional keyword param. If raw is False (which is the 
                     default) it will try to instantiated the element to the 
                     element_class specified when the proxy object was created.
                     However, if the raw param is set to True, it will won't 
                     instantiate the object and will return the raw Response 
                     object.
        """
        target = "%s/%s" % (self._base_target(),_id)
        resp = self.resource.post(target,data)
        if raw is True:
            return resp
        return self.element_class(self.resource,resp.results)

    def get(self,_id):
        """ 
        Retrieves an element from Rexster and returns it. 
        
        :param _id: The ID of the element you want to retrieve.

        """
        target = "%s/%s" % (self._base_target(),_id)
        resp = self.resource.get(target,params=None)
        if resp.results:
            return self.element_class(self.resource,resp.results)        
    
    def get_all(self):
        """Returns all the elements from the Rexster of _type vertex or edge."""
        resp = self.resource.get(self._base_target(),params=None)
        for result in resp.results:
            yield self.element_class(self.resource,result)
            
    def remove(self,_id,params):
        """ 
        Removes properties from a vertex and returns the response.

        :param _id: The ID of the element.

        :param params: The properties you want to remove.
        """ 
        # shouldn't this return the updated vertex instead?
        if params is None:
            raise Exception("params are required.")
        target = "%s/%s" % (self._base_target(),_id)
        resp = self.resource.delete(target,params)
        return resp

    def delete(self,element):
        """ 
        Deletes a vertex from a graph DB and returns the response.
        
        :param element: The element you want to delete.
        """
        # TODO: Why are we requiring the full element and not just the ID?
        target = "%s/%s" % (self._base_target(),element._id)
        resp = self.resource.delete(target,params=None)
        return resp


class VertexProxy(ElementProxy):
    """ A proxy for interacting with vertices on Rexster. """

    def __init__(self,resource,element_class=Vertex):
        """
        Initializes the proxy object.

        :param resource: The Resource object for the database.

        :param element_class: A subclass of Vertex used to instantiate objects.

        """
        assert issubclass(element_class,Vertex)
        ElementProxy.__init__(self,resource,element_class)

    @classmethod
    def _path(self):
        """Returns Rexster's URL path segment used for vertices."""
        return "vertices"


class EdgeProxy(ElementProxy):
    """ A proxy for interacting with edges on Rexster. """

    def __init__(self,resource,element_class=Edge):
        """
        :param resource: The Resource object for the database.

        :param element_class: A subclass of Edge used to instantiate objects.

        """
        assert issubclass(element_class,Edge)
        ElementProxy.__init__(self,resource,element_class)

    def create(self,outV,label,inV,data={},raw=False):
        """ 
        Adds an edge to the database and returns it. 

        :param outV: The outgoing vertex. You can pass the vertex in as either
                      an ID or a Vertex object, and it will automatically 
                      convert it to an ID for you.
 
        :param label: The edge's label.

        :param inV: The incoming vertex. You can pass the vertex in as either 
                     an ID or a Vertex object, and it will automatically 
                     convert it to an ID for you.

        :param data: The edge's property data to save in the database.
        
        :param raw: Optional keyword param. If raw is False (which is the 
                     default) it will try to instantiated the element to the 
                     element_class specified when the proxy object was created.
                     However, if the raw param is set to True, it will won't 
                     instantiate the object and will return the raw Response 
                     object.
        """
        assert label is not None
        outV = self._coerce_vertex_id(outV)
        inV = self._coerce_vertex_id(inV)
        edge_data = dict(_outV=outV,_label=label,_inV=inV)
        data.update(edge_data)
        return ElementProxy.create(self,data,raw)

    def _path(self):
        """Returns Rexster's URL path segment used for edges."""
        return "edges"
    
    def _coerce_vertex_id(self,v):
        """ 
        Returns the vertex ID coerced into an int if need be.

        :param v: Either a Vertex object or a vertex ID.

        """
        if isinstance(v,Vertex):
            vertex_id = v._id
        else:
            # the vertex ID may have been passed in as a string
            #vertex_id = int(v)
            vertex_id = utils.coerce_id(v)
        return vertex_id

  



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
import unittest

from bulbs.graph import Graph
       
class IndexTestCase(unittest.TestCase):
    
    def test_index(self):
        self.graph = Graph()
        index_name = "TEST"
        # need to fix this to accept actual data types in POST
        ikeys = '[name,location]'
        self.graph.indices.delete(index_name)
        i1 = self.graph.indices.create(index_name,"vertex","automatic",ikeys)
        assert i1.index_name == index_name
        assert i1.index_type == "automatic"
        james = self.graph.vertices.create({'name':'James'})
        i1.put('name','James',james._id)
        i1.put('location','Dallas',james._id)
        results = i1.get('name','James')
        results = list(results)
        assert len(results) == 1
        assert results[0].name == "James"
        total_size = i1.count('name','James')
        assert total_size == 1
        keys = i1.keys()
        assert 'name' in keys
        assert 'location' in keys
        i2 = self.graph.indices.get(index_name)
        print "INDEX_NAME", index_name, i1.index_name, i2.index_name
        assert i1.index_name == i2.index_name
        i1.remove(james._id,'name','James')
        james = i1.get_unique('name','James')
        assert james is None
        results = i1.rebuild()
        assert type(results) == list
        self.graph.indices.delete(index_name)

        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(IndexTestCase))

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




import unittest
from bulbs import config
from bulbs.graph import Graph

class GraphTestCase(unittest.TestCase):

    def setUp(self):
        self.graph = Graph()

    def test_init(self):
        assert config.DATABASE_URL == self.graph.resource.db_url

    def test_V(self):
        vertices = self.graph.V
        vertices = list(vertices)
        assert len(vertices) > 0
        
    def test_E(self):
        edges = self.graph.E
        edges = list(edges)
        assert len(edges) > 0

    def test_idxV(self):
        self.graph.vertices.create({'name':'james'})
        vertices = self.graph.idxV(name="james")
        vertices = list(vertices)
        assert len(vertices) > 0
        
    def test_idxE(self):
        self.james = self.graph.vertices.create({'name':'James'})
        self.julie = self.graph.vertices.create({'name':'Julie'})
        self.graph.edges.create(self.james,"test",self.julie,{'time':'rightnow'})
        edges = self.graph.idxE(time="rightnow")
        edges = list(edges)
        assert len(edges) > 0

    #def test_clear(self):
    #    # WARNING: Be careful about uncommenting this. It will wipe your data.
    #    self.graph.clear()
    #    vertices = self.graph.V
    #    vertices = list(vertices)
    #    assert len(vertices) == 0
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(GraphTestCase))

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
    Bulbs is a Python persistence framework for graph databases.
    
    Bulbs is a Python client library for Rexster, a REST server that 
    provides access to any Blueprints-enabled graph database, including 
    TinkerGraph, Neo4j, OrientDB, DEX, and Sail RDF Stores. 

    Home Page: http://bulbflow.com
 
    Copyright (c) 2011, James Thornton (http://jamesthornton.com)
    License: BSD (see LICENSE for details)
"""
__version__ = '0.2.1'
__author__ = 'James Thornton'
__license__ = 'BSD'



# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Interface for interacting with a graph database through Rexster.

Graph, Vertex, Edge, Index, Query

"""
#import array


def convert_to_python(data):
    datatype = data['type']
    value = data['value']
    klass = type_map[datatype]
    return klass.to_python(value)

def convert_to_rexster(data):
    datatype = type(data)
    value = data
    klass = type_map[datatype]
    return klass.to_rexster(value)


# NOTE 1: "Property" refers to a graph-database property (i.e. the DB data)
class Property(object):

    def __init__(self, datatype, fget=None, fset=None, fdel=None, \
                     name=None, default=None, onupdate=None, constraint=None, \
                     nullable=True, unique=False, index=False):

        # TODO: unique creates an index

        #args = list(args)
        #name = args.pop(0)
        #proptype = args.pop(0)

        #self.nullable = kwargs.pop('nullable', not self.primary_key)
        #self.default = kwargs.pop('default', None)
        #self.index = kwargs.pop('index', None)
        #self.unique = kwargs.pop('unique', None)
        self.datatype = datatype
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        # NOTE: If you pass name as a kwd, then it overwrites variables named "name"
        # FIX THIS!!!! -- why are they sharing the same namespace???
        self.name = name
        self.default = default
        self.onupdate = onupdate
        self.constraint = constraint,
        self.nullable = nullable
        self.unique = unique
        self.index = index

    def is_valid(self, value):
        return isinstance(value, self.datatype.python_type)


class DataType(object):

    def __init__(self):
        # TODO: Is this being used or 
        # are you explicitly setting them at bottom of this module?
        type_map.update(self.python_type,self.__class__)
        type_map.update(self.rexter_type,self.__class__)

    @classmethod
    def convert(self,value,datatype):
        # TODO: warn or log errors
        try: return datatype(value)
        except: return None


class String(DataType): 
    python_type = str
    rexster_type = "string"

    @classmethod
    def to_rexster(self,value):
        return '(string,%s)' % (value)
    
    @classmethod
    def to_python(self,value):
        return self.convert(value,str)

class Integer(DataType):    
    python_type = int
    rexter_type = "integer"
    
    @classmethod
    def to_rexster(self,value):
        return '(integer,%d)' % (value)
    
    @classmethod
    def to_python(self,value):
        return self.convert(value,int)

class Long(DataType):
    python_type = long
    rexster_type = "long"
    
    @classmethod
    def to_rexster(self,value):
        return '(long,%d)' % (value)

    @classmethod
    def to_python(self,value):
        return self.convert(value,long)

class Float(DataType):
    python_type = float
    rexter_type = "float"
    
    @classmethod
    def to_rexster(self,value):
        return '(float,%f)' % (value)
    
    @classmethod
    def to_python(self,value):
        return self.convert(value,float)              

class Null(DataType):
    python_type = None
    rexter_type = "string"

    @classmethod
    def to_rexster(self,value):
        # TODO: we prob don't want to save None values as 
        # empty strings in Rexter, instead just ignore them
        return '(string,%s)' % ('')
    
    @classmethod
    def to_python(self,value):
        return None

class List(DataType):
    python_type = list
    rexster_type = "list"

    @classmethod
    def to_rexster(self,value):
        val_list = []
        for item in value:
            val_list.append(convert_to_rexster(item))
        return "(list,(%s))" % (','.join(val_list))    

    @classmethod
    def to_python(self,value):
        val_list = []
        for item in value:
            val_list.append(convert_to_python(item))
        return val_list

class Dictionary(DataType):
    python_type = dict
    rexster_type = "map"

    @classmethod
    def to_rexster(self,value):
        pair_list = []
        for k,v in value.items():
            pair = "%s=%s" % (k,convert_to_rexster(v))
            pair_list.append(pair)
        return "(map,(%s))" % (','.join(pair_list))    

    @classmethod
    def to_python(self,value):
        dict_ = {}
        for k,v in value.items():
            dict_.update({k:convert_to_python(v)})
        return dict_

type_map = dict()
type_map[str] = String
type_map['string'] = String
type_map[int] = Integer
type_map['integer'] = Integer
type_map[float] = Float
type_map['float'] = Float
type_map[list] = List
type_map['list'] = List
type_map[dict] = Dictionary
type_map['map'] = Dictionary
type_map['array'] = List
type_map[type(None)] = Null
